// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//
// ndconn.h - NetworkDirect connection scalability test
//

#pragma once


class __declspec( novtable) CReference
{
public:
    CReference() : m_nRef( 1 ){};

    void AddRef()
    {
        InterlockedIncrement( &m_nRef );
    }

    void Release()
    {
        if( InterlockedDecrement( &m_nRef ) > 0 )
        {
            return;
        }

        delete this;
    }

protected:
    virtual ~CReference(){};

private:
    LONG m_nRef;
};


class COverlapped : public OVERLAPPED
{
public:
    typedef void (*CompletionRoutine)( _In_ COverlapped* This );

public:
    COverlapped( _In_ CompletionRoutine pfnSucceeded, _In_ CompletionRoutine pfnFailed ) :
        m_pfnSucceeded( pfnSucceeded ),
        m_pfnFailed( pfnFailed )
    {
        Internal = 0;
        InternalHigh = 0;
        Pointer = NULL;
        hEvent = NULL;
    };

    void Succeeded(){ m_pfnSucceeded( this ); }
    void Failed(){ m_pfnFailed( this ); }

protected:
    CompletionRoutine m_pfnSucceeded;
    CompletionRoutine m_pfnFailed;
};


class CTest
{
public:
    CTest( _In_ SIZE_T BufferSize );
    ~CTest();

protected:
    void Start();

protected:
    HANDLE m_hIocp;
    INDAdapter* m_pAdapter;
    char* m_pBuf;
    SIZE_T m_BufferSize;
    ND_MR_HANDLE m_hMr;

    volatile bool m_fEndTest;
};


//
// Client test object.
//
class CClientTest : public CTest
{
    friend class CClientEp;

public:
    CClientTest( _In_ SIZE_T BufferSize, _In_ LONG nThreads );
    ~CClientTest();

    void Run();
    void OpenAdapter( _In_ const struct sockaddr_in& v4 );

protected:
    __callback static DWORD CALLBACK TestRoutine( _In_ LPVOID This );

private:
    __callback static void SendSucceeded( _In_ COverlapped* pOv );
    __callback static void SendFailed( _In_ COverlapped* pOv );
    __callback static void RecvSucceeded( _In_ COverlapped* pOv );
    __callback static void RecvFailed( _In_ COverlapped* pOv );

protected:
    INDCompletionQueue* m_pSendCq;
    COverlapped m_SendOv;
    INDCompletionQueue* m_pRecvCq;
    COverlapped m_RecvOv;

    struct sockaddr_in m_Addr;

    LONG m_nThreads;

    volatile LONG m_nEpCreated;
    volatile LONG m_nEpDestroyed;

    volatile LONGLONG m_ConnectTime;
    volatile LONGLONG m_CompleteConnectTime;
    volatile LONGLONG m_DisconnectTime;

    volatile LONG m_nConnFailure;
    volatile LONG m_nConnTimeout;
};


//
// Server test object.
//
class CServerTest : public CTest
{
    friend class CServerEp;
    friend class CConnReq;

public:
    CServerTest( _In_ SIZE_T BufferSize, _In_ LONG nThreads );
    ~CServerTest();

    void Run();
    void OpenAdapter( _In_ const struct sockaddr_in& v4 );

protected:
    __callback static DWORD CALLBACK TestRoutine( _In_ LPVOID This );

private:
    __callback static void SendSucceeded( _In_ COverlapped* pOv );
    __callback static void SendFailed( _In_ COverlapped* pOv );
    __callback static void RecvSucceeded( _In_ COverlapped* pOv );
    __callback static void RecvFailed( _In_ COverlapped* pOv );

protected:
    INDListen* m_pListen;
    INDCompletionQueue* m_pSendCq;
    COverlapped m_SendOv;
    INDCompletionQueue* m_pRecvCq;
    COverlapped m_RecvOv;

    volatile LONGLONG m_AcceptTime;
    volatile LONGLONG m_DisconnectTime;

    LONG m_nThreads;

    volatile LONG m_nEpCreated;
    volatile LONG m_nEpDestroyed;
    volatile LONG m_nOv;
    volatile LONG m_nConnFailure;
};


class CConnReq : public COverlapped
{
public:
    CConnReq( CServerTest* pTest ) :
        COverlapped( GetConnSucceeded, GetConnFailed ),
        m_pTest( pTest ),
        m_pConnector( NULL )
    {
        GetNextRequest();
    };

    ~CConnReq()
    {
        if( m_pConnector )
        {
            m_pConnector->Release();
        }
    }

private:
    void GetNextRequest();

private:
    __callback static void GetConnSucceeded( _In_ COverlapped* pOv );
    __callback static void GetConnFailed( _In_ COverlapped* pOv );

public:
    CServerTest* m_pTest;
    INDConnector* m_pConnector;
};


class CServerEp : public CReference
{
    friend class CServerTest;

private:
    CServerEp( _In_ CServerTest* pTest );
    ~CServerEp();

    HRESULT Init( _In_ INDConnector* pConnector );

public:
    static HRESULT Create( _In_ CServerTest* pTest, _In_ INDConnector* pConnector );

private:
    void Send();
    void AcceptError( _In_ HRESULT hr );

private:
    __callback static void AcceptSucceeded( _In_ COverlapped* pOv );
    __callback static void AcceptFailed( _In_ COverlapped* pOv );
    __callback static void DisconnectSucceeded( _In_ COverlapped* pOv );
    __callback static void DisconnectFailed( _In_ COverlapped* pOv );
    __callback static void NotifyDisconnectSucceeded( _In_ COverlapped* pOv );
    __callback static void NotifyDisconnectFailed( _In_ COverlapped* pOv );

    static void SendDone( _In_ ND_RESULT* pResult );
    static void RecvDone( _In_ ND_RESULT* pResult );

private:
    Timer m_Timer;
    CServerTest* m_pTest;
    INDConnector* m_pConnector;
    INDEndpoint* m_pEp;
    COverlapped m_AcceptOv;
    COverlapped m_DisconnectOv;
    COverlapped m_NotifyDisconnectOv;
    ND_RESULT m_Result;

    volatile LONG m_fDoSend;
};


class CClientEp : public CReference
{
    friend class CClientTest;

private:
    CClientEp( _In_ CClientTest* pTest );
    ~CClientEp();

    HRESULT Init();

public:
    static HRESULT Create( _In_ CClientTest* pTest );

private:
    void ConnectError( _In_ HRESULT hr );

private:
    __callback static void ConnectSucceeded( _In_ COverlapped* pOv );
    __callback static void ConnectFailed( _In_ COverlapped* pOv );
    __callback static void CompleteConnectSucceeded( _In_ COverlapped* pOv );
    __callback static void CompleteConnectFailed( _In_ COverlapped* pOv );
    __callback static void DisconnectSucceeded( _In_ COverlapped* pOv );
    __callback static void DisconnectFailed( _In_ COverlapped* pOv );

    static void SendDone( _In_ ND_RESULT* pResult );
    static void RecvDone( _In_ ND_RESULT* pResult );

private:
    Timer m_Timer;
    CClientTest* m_pTest;
    INDConnector* m_pConnector;
    INDEndpoint* m_pEp;
    COverlapped m_ConnectOv;
    COverlapped m_CompleteConnectOv;
    COverlapped m_DisconnectOv;
    ND_RESULT m_Result;
};


