// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//
// ndmrlat.cpp - NetworkDirect memory registration bandwidth test
//

#include "ndcommon.h"


const SIZE_T x_MaxSize = (4 * 1024 * 1024);
char x_Buf[x_MaxSize] = {0};
const LONG x_MaxReg = 10000;
const LONG x_MinReg = 2000; // Number of iterations for x_MaxSize.

struct MR_CONTEXT
{
    OVERLAPPED Ov;
    ND_MR_HANDLE hMr;
};

MR_CONTEXT g_MrContext[x_MaxReg] = {0};

INDAdapter* g_pAdapter = NULL;
HANDLE g_hIocp = NULL;
HANDLE g_hEvent = NULL;

volatile LONG g_nReg = 0;
volatile LONG g_nMr = 0;

//
// Registers buffers of increasing size 10000 times using multiple threads.
//


void ShowUsage()
{
    printf( "ndmrrate <ip> t<threads> [l<log>]\n"
        "\t<ip> - IPv4 Address\n"
        "\tt - number of threads\n"
        "\tl - log output to a file named <log>.\n");
}


DWORD WINAPI TestRoutine( _In_opt_ LPCVOID Nothing )
{
    UNREFERENCED_PARAMETER( Nothing );

    Timer timer;
    CpuMonitor cpu;
    LONG iterations;

    for( SIZE_T szXfer = 1; szXfer <= x_MaxSize; szXfer <<= 1 )
    {
        if( (x_MaxSize / szXfer) > (x_MaxReg / x_MinReg) )
        {
            iterations = x_MaxReg;
        }
        else
        {
            iterations = static_cast<ULONG>( (x_MaxSize / szXfer) * x_MinReg );
        }

        bool Register = true;
        g_nMr = 0;
        g_nReg = 0;

        cpu.Start();
        timer.Start();
        while( g_nMr != iterations )
        {
            while( g_nReg != iterations )
            {
                //
                // Register
                //
                LONG iReg = InterlockedIncrement( &g_nReg ) - 1;
                if( iReg >= iterations )
                {
                    InterlockedDecrement( &g_nReg );
                    break;
                }

                if( Register )
                {
                    HRESULT hr = g_pAdapter->RegisterMemory(
                        x_Buf,
                        szXfer,
                        &g_MrContext[iReg].Ov,
                        &g_MrContext[iReg].hMr );

                    if( FAILED( hr ) )
                    {
                        printf( "INDAdapter::RegisterMemory failed with %08x\n", hr );
                        exit( __LINE__ );
                    }
                }
                else
                {
                    HRESULT hr = g_pAdapter->DeregisterMemory(
                        g_MrContext[iReg].hMr,
                        &g_MrContext[iReg].Ov );

                    if( FAILED( hr ) )
                    {
                        printf( "INDAdapter::DeregisterMemory failed with %08x\n", hr );
                        exit( __LINE__ );
                    }
                }
            }
            //
            // Check the IOCP for completions.
            //
            DWORD BytesRet;
            ULONG_PTR Key;
            OVERLAPPED* pOv;
            BOOL fSuccess = GetQueuedCompletionStatus(
                g_hIocp,
                &BytesRet,
                &Key,
                &pOv,
                INFINITE
                );

            if( fSuccess == FALSE )
            {
                printf( "GetQueuedCompletionStatus failed with %u\n", GetLastError() );
                exit( __LINE__ );
            }

            SIZE_T BytesRet2;
            HRESULT hr = g_pAdapter->GetOverlappedResult( pOv, &BytesRet2, FALSE );
            if( FAILED( hr ) )
            {
                printf( "INDAdapter->GetOverlappedResult failed with %08x\n", hr );
                exit( __LINE__ );
            }

            InterlockedIncrement( &g_nMr );
            if( g_nMr == iterations && Register )
            {
                timer.Split();
                cpu.Split();
                //
                // Repeat the loop for deregistration.
                //
                Register = false;
                g_nMr = 0;
                g_nReg = 0;
            }
        }

        timer.End();
        cpu.End();
        printf(
            "%9Id %9d %9.2f %7.2f %9.2f %7.2f\n",
            szXfer,
            iterations,
            timer.ReportPreSplit() / iterations,
            cpu.ReportPreSplit(),
            timer.ReportPostSplit() / iterations,
            cpu.ReportPostSplit()
            );
    }
    return 0;
}


int __cdecl _tmain(int argc, TCHAR* argv[])
{
    struct sockaddr_in v4 = {0};

    WSADATA wsaData;
    int ret = ::WSAStartup( MAKEWORD( 2, 2 ), &wsaData );
    if( ret != 0 )
    {
        printf( "Failed to initialize Windows Sockets: %d\n", ret );
        exit( __LINE__ );
    }

    DWORD nThreads = 1;

    for( int i = 1; i < argc; i++ )
    {
        const TCHAR* pArg;
        INT len;

        pArg = argv[i];

        // Skip leading dashes/slashes
        while( *pArg == '-' || *pArg == '/' )
        {
            pArg++;
        }

        switch( *pArg )
        {
        case _T('t'):
        case _T('T'):
            nThreads = _ttol( ++pArg );
            break;
        case _T('l'):
        case _T('L'):
//
// Can't use freopen_s because it doesn't allow sharing.
// So supress the deprecated warning, because for our use
// it isn't deprecated.
//
#pragma warning( disable : 4996 )
            if( _tfreopen( ++pArg, _T("w"), stdout ) == NULL ||
                _tfreopen( pArg, _T("a+"), stderr ) == NULL )
#pragma warning( default : 4996 )
            {
                printf( "Could not open log file.\n" );
                exit( __LINE__ );
            }
            break;
        default:
            len = sizeof( v4 );
            WSAStringToAddress(
                argv[i],
                AF_INET,
                NULL,
                reinterpret_cast<struct sockaddr*>( &v4 ),
                &len
                );
            break;
        }
    }

    if( v4.sin_addr.s_addr == 0 )
    {
        printf( "Bad address.\n");
        ShowUsage();
        exit( __LINE__ );
    }

    if( nThreads == 0 )
    {
        printf( "Invalid number of threads.\n" );
        ShowUsage();
        exit( __LINE__ );
    }

    HRESULT hr = NdStartup();
    if( FAILED(hr) )
    {
        printf( "NdStartup failed with %08x\n", hr );
        exit( __LINE__ );
    }

    g_hEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
    if( g_hEvent == NULL )
    {
        printf( "CreateEvent failed with %u\n", GetLastError() );
        exit( __LINE__ );
    }

    hr = NdOpenV1Adapter( (const struct sockaddr*)&v4, sizeof(v4), &g_pAdapter );
    if( FAILED( hr ) )
    {
        printf( "NdOpenV1Adapter failed with %08x\n", hr );
        exit( __LINE__ );
    }

    g_hIocp = CreateIoCompletionPort( g_pAdapter->GetFileHandle(), NULL, 0, 0 );
    if( g_hIocp == NULL )
    {
        printf( "Failed to bind adapter to IOCP, error %u\n", GetLastError() );
        exit( __LINE__ );
    }

    printf(
        "Using %u processors. Sender Frequency is %I64d\n",
        CpuMonitor::CpuCount(),
        Timer::Frequency()
        );

    printf(
        "                         Register          Deregister\n"
        "     Size     Iter    usec      cpu       usec      cpu\n"
        );

    HANDLE* hThreads = new (std::nothrow) HANDLE[nThreads];
    if( hThreads == NULL )
    {
        printf( "Failed to allocate thread handle array\n" );
        exit( __LINE__ );
    }

    //for( DWORD i = 0; i < nThreads; i++ )
    //{
    //    hThreads[i] = CreateThread( NULL, 0, TestRoutine, NULL, 0, 0 );
    //    if( hThreads[i] == NULL )
    //    {
    //        printf( "CreateThread for thread %d of %d failed with %d.\n", i + 1, nThreads, GetLastError() );
    //        exit( __LINE__ );
    //    }
    //}

    ////
    //// Wait for the threads to exit.
    ////
    //for( DWORD i = 0; i < nThreads; i++ )
    //{
    //    WaitForSingleObject( hThreads[i], INFINITE );
    //    CloseHandle( hThreads[i] );
    //    hThreads[i] = NULL;
    //}
    TestRoutine( NULL );

    g_pAdapter->Release();

    CloseHandle( g_hIocp );

    hr = NdCleanup();
    if( FAILED(hr) )
    {
        printf( "NdCleanup failed with %08x\n", hr );
        exit( __LINE__ );
    }

    delete[] hThreads;

    _fcloseall();
    WSACleanup();

    return 0;
}

