// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//
// ndpingpong.cpp - NetworkDirect send/recv pingpong test
//

#include "ndcommon.h"
#include "ndshim.h"

const USHORT x_DefaultPort = 54325;
const DWORD x_MaxXfer = (4 * 1024 * 1024);
const DWORD x_HdrLen = 40;
const SIZE_T x_MaxVolume = (500 * x_MaxXfer);
const DWORD x_MaxIterations = 100000;


#define RECV    0
#define SEND    1


void ShowUsage()
{
    printf( "ndpingpong s|c <ip>[:<port>] [b|p<nSge>] [l<log>]\n"
        "\ts - start as server (listen on IP/Port)\n"
        "\tc - start as client (connect to server IP/Port)\n"
        "\t<ip> - IPv4 Address\n"
        "\t<port> - Port number, default %hu\n"
        "\tb - blocking I/O (wait for CQ notification)\n"
        "\tp - polling I/O (Poll on the CQ) (default)\n"
        "\t<nSge> - Number of scatter/gather entries per transfer, 1 or 2, default 1.\n"
        "\tv - ND provider version, default 1.\n"
        "\tq - Receive queue depth, default 64.\n"
        "\tl - log output to a file named <log>.\n",
        x_DefaultPort
        );
}


class NdPingPongBase
{
protected:
    IND2Adapter* m_pAdapter;
    IND2MemoryRegion* m_pMr;
    IND2CompletionQueue* m_pCq;
    IND2QueuePair* m_pQp;
    IND2Connector* m_pConnector;

    HANDLE m_hAdapterFile;

    char* m_Buf;

    ND_RESULT m_Results[2];

    OVERLAPPED m_Ov;

protected:
    NdPingPongBase();
    ~NdPingPongBase();

    void Init(
        _In_ IND2Adapter* pAdapter,
        _In_ DWORD queueDepth,
        _In_ DWORD nSge
        );
    void Shutdown();
};

NdPingPongBase::NdPingPongBase() :
    m_pAdapter( NULL ),
    m_pMr( NULL ),
    m_pCq( NULL ),
    m_pQp( NULL ),
    m_pConnector( NULL ),
    m_hAdapterFile( NULL ),
    m_Buf( NULL )
{
    RtlZeroMemory( &m_Ov, sizeof(m_Ov ) );
}

NdPingPongBase::~NdPingPongBase()
{
    if( m_pMr != NULL )
    {
        m_pMr->Release();
    }

    if( m_pCq != NULL )
    {
        m_pCq->Release();
    }

    if( m_pQp != NULL )
    {
        m_pQp->Release();
    }

    if( m_pConnector != NULL )
    {
        m_pConnector->Release();
    }

    if( m_hAdapterFile != NULL )
    {
        CloseHandle( m_hAdapterFile );
    }

    if( m_pAdapter != NULL )
    {
        m_pAdapter->Release();
    }

    if( m_Ov.hEvent != NULL )
    {
        CloseHandle( m_Ov.hEvent );
    }

    if( m_Buf != NULL )
    {
        delete[] m_Buf;
    }
}

void NdPingPongBase::Init(
    _In_ IND2Adapter* pAdapter,
    _In_ DWORD queueDepth,
    _In_ DWORD nSge
    )
{
    pAdapter->AddRef();
    m_pAdapter = pAdapter;

    m_Ov.hEvent = CreateEvent( NULL, FALSE, FALSE, NULL );
    if( m_Ov.hEvent == NULL )
    {
        printf( "Failed to allocate event for overlapped operations.\n" );
        exit( __LINE__ );
    }

    //
    // Get the file handle for overlapped operations on this adapter.
    //
    HRESULT hr = m_pAdapter->CreateOverlappedFile( &m_hAdapterFile );
    if( FAILED( hr ) )
    {
        printf( "IND2Adapter::CreateOverlappedFile failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Allocate and register the data buffer.
    //
    m_Buf = new (std::nothrow) char[x_HdrLen + x_MaxXfer];
    if( m_Buf == NULL )
    {
        printf( "Failed to allocate buffer.\n" );
        exit( __LINE__ );
    }
    hr = m_pAdapter->CreateMemoryRegion(
        IID_IND2MemoryRegion,
        m_hAdapterFile,
        reinterpret_cast<VOID**>( &m_pMr )
        );
    if( FAILED( hr ) )
    {
        printf( "IND2Adapter::CreateMemoryRegion failed with %08x\n", hr );
        exit( __LINE__ );
    }

    hr = m_pMr->Register(
        m_Buf,
        x_HdrLen + x_MaxXfer,
        ND_MR_FLAG_ALLOW_LOCAL_WRITE,
        &m_Ov
        );
    if( hr == ND_PENDING )
    {
        hr = m_pMr->GetOverlappedResult( &m_Ov, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "IND2MemoryRegion::Register failed with %08x\n", hr );
        exit( __LINE__ );
    }

    //
    // Create a CQ.
    //
    hr = m_pAdapter->CreateCompletionQueue(
        IID_IND2CompletionQueue,
        m_hAdapterFile,
        queueDepth + 1,
        0,
        0,
        reinterpret_cast<VOID**>( &m_pCq )
        );
    if( FAILED( hr ) )
    {
        printf( "IND2Adapter::CreateCompletionQueue failed with %08x\n", hr );
        exit( __LINE__ );
    }

    hr = m_pAdapter->CreateConnector(
        IID_IND2Connector,
        m_hAdapterFile,
        reinterpret_cast<VOID**>( &m_pConnector )
        );
    if( FAILED( hr ) )
    {
        printf( "IND2Adapter::CreateConnector failed with %08x", hr );
        exit( __LINE__ );
    }

    hr = m_pAdapter->CreateQueuePair(
        IID_IND2QueuePair,
        m_pCq,
        m_pCq,
        NULL,
        queueDepth,
        1,
        nSge,
        nSge,
        160,
        reinterpret_cast<VOID**>( &m_pQp )
        );
    if( FAILED( hr ) )
    {
        printf( "IND2Connector::CreateEndpoint failed with %08x\n", hr );
        exit( __LINE__ );
    }
}

void NdPingPongBase::Shutdown()
{
    //
    // Test is complete.  Note that during teardown we don't treat
    // errors as test failures, but we still report them.
    //
    HRESULT hr = m_pConnector->Disconnect( &m_Ov );
    if( hr == ND_PENDING )
    {
        hr = m_pConnector->GetOverlappedResult( &m_Ov, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "IND2QueuePair::Disconnect failed wtih %08x\n", hr );
    }

    hr = m_pMr->Deregister( &m_Ov );
    if( hr == ND_PENDING )
    {
        hr = m_pMr->GetOverlappedResult( &m_Ov, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "IND2Adapter::DeregisterMemory failed with %08x\n", hr );
    }
}


class NdPingPongServerBase : public NdPingPongBase
{
protected:
    IND2Listener* m_pListener;

public:
    NdPingPongServerBase();
    ~NdPingPongServerBase();

    void GetConnection( _In_ const sockaddr_in& v4Src );
    void Accept();
};

NdPingPongServerBase::NdPingPongServerBase() :
    m_pListener( NULL )
{
}

NdPingPongServerBase::~NdPingPongServerBase()
{
    if( m_pListener != NULL )
    {
        m_pListener->Release();
    }
}

void NdPingPongServerBase::GetConnection( _In_ const sockaddr_in& v4Src )
{
    HRESULT hr = m_pAdapter->CreateListener(
        IID_IND2Listener,
        m_hAdapterFile,
        reinterpret_cast<VOID**>( &m_pListener )
        );
    if( FAILED( hr ) )
    {
        printf( "IND2Adapter::CreateListener failed with %08x\n", hr );
        exit( __LINE__ );
    }

    hr = m_pListener->Bind(
        reinterpret_cast<const sockaddr*>( &v4Src ),
        sizeof( v4Src )
        );
    if( FAILED( hr ) )
    {
        printf( "IND2Listener::Bind failed with %08x\n", hr );
        exit( __LINE__ );
    }

    hr = m_pListener->Listen( 0 );
    if( FAILED( hr ) )
    {
        printf( "IND2Listen::Listener failed with %08x\n", hr );
        exit( __LINE__ );
    }

    hr = m_pListener->GetConnectionRequest( m_pConnector, &m_Ov );
    if( hr == ND_PENDING )
    {
        hr = m_pListener->GetOverlappedResult( &m_Ov, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "IND2Listen::GetConnectionRequest failed with %08x\n", hr );
        exit( __LINE__ );
    }
}

void NdPingPongServerBase::Accept()
{
    //
    // Accept the connection.
    //
    HRESULT hr = m_pConnector->Accept(
        m_pQp,
        0,
        0,
        NULL,
        0,
        &m_Ov
        );
    if( hr == ND_PENDING )
    {
        hr = m_pConnector->GetOverlappedResult( &m_Ov, TRUE );
    }
    if( FAILED( hr ) )
    {
        printf( "IND2Connector::Accept failed with %08x\n", hr );
        exit( __LINE__ );
    }
}


template<class T>
class NdPingPongServer : public NdPingPongServerBase
{
    T m_IoHelper;

public:
    void RunTest(
        _In_ const struct sockaddr_in& v4Src,
        _In_ DWORD queueDepth,
        _In_ DWORD nSge,
        _In_ bool bUseEvents
        );

    void Pong(
        _In_ DWORD nSge,
        _In_ DWORD iterations,
        _In_ bool bUseEvents
        );
};

template<class T>
void NdPingPongServer<T>::RunTest(
    _In_ const struct sockaddr_in& v4Src,
    _In_ DWORD queueDepth,
    _In_ DWORD nSge,
    _In_ bool bUseEvents
    )
{
    IND2Adapter* pAdapter;
    T::OpenAdapter( v4Src, &pAdapter, __LINE__ );

    NdPingPongBase::Init( pAdapter, queueDepth, nSge );
    pAdapter->Release();

    m_IoHelper.Init( m_pQp, m_pCq, m_pMr, m_Buf, x_HdrLen + x_MaxXfer, nSge, __LINE__ );

    NdPingPongServerBase::GetConnection( v4Src );

    //
    // Pre-post receive request.
    //
    for( SIZE_T i = 0; i < queueDepth; i++ )
    {
        m_IoHelper.Receive( &m_Results[RECV], __LINE__ );
    }

    NdPingPongServerBase::Accept();

    //
    // When using the SHIM, the INDEndpoint object in our IND2QueuePair object
    // changes in the Accept call.  Refresh our reference too.
    //
    m_IoHelper.RefreshQp( m_pQp, __LINE__ );

    //
    // Warm up run, transfer only header.
    //
    m_IoHelper.SetSendSge( 0, m_Buf, x_HdrLen );

    Pong(
        nSge,
        1000,
        bUseEvents
        );

    for( DWORD szXfer = 1; szXfer <= x_MaxXfer; szXfer <<= 1 )
    {
        DWORD i;
        for( i = 0; i < nSge - 1; i++ )
        {
            m_IoHelper.SetSendSge( i, m_Buf + (x_HdrLen * i), x_HdrLen );
        }

        // Last SGE has the remainder of the data.
        m_IoHelper.SetSendSge( i, m_Buf + (x_HdrLen * i), x_HdrLen + szXfer - (x_HdrLen * i) );

        DWORD iterations = x_MaxIterations;
        if( iterations > (x_MaxVolume / szXfer) )
        {
            iterations = x_MaxVolume / szXfer;
        }

        Pong(
            nSge,
            iterations,
            bUseEvents
            );
    }

    NdPingPongBase::Shutdown();
}

template<class T>
void NdPingPongServer<T>::Pong(
    _In_ DWORD nSge,
    _In_ DWORD iterations,
    _In_ bool bUseEvents
    )
{
    DWORD cbExpected = 0;
    for( DWORD i = 0; i < nSge; i++ )
    {
        cbExpected += m_IoHelper.GetSendSgeLength(i);
    }
    ULONG flags = 0;
    if( cbExpected < m_IoHelper.MaxInline() )
    {
        flags |= ND_OP_FLAG_INLINE;
    }

    void* context;
    DWORD cbXfer;
    do
    {
        //
        // Wait for receive completion.
        //
        if( m_IoHelper.GetResults( &context, &cbXfer, __LINE__ ) == 0 )
        {
            if( bUseEvents )
            {
                HRESULT hr = m_pCq->Notify( ND_CQ_NOTIFY_ANY, &m_Ov );
                if( hr == ND_PENDING )
                {
                    hr = m_pCq->GetOverlappedResult( &m_Ov, TRUE );
                }
                if( FAILED( hr ) )
                {
                    printf( "IND2CompletionQueue::Notify failed with %08x\n", hr );
                    exit( __LINE__ );
                }
            }
            while( m_IoHelper.GetResults( &context, &cbXfer, __LINE__ ) == 0 )
            {
                NULL;
            }
        }

        if( context == &m_Results[RECV] && cbXfer != cbExpected )
        {
            printf( "Received %u bytes, expected %u\n",
                cbXfer, cbExpected );
            exit( __LINE__ );
        }

        //
        // Repost the receive.
        //
        m_IoHelper.Receive( &m_Results[RECV], __LINE__ );
        m_IoHelper.Send( &m_Results[SEND], nSge, __LINE__, flags );

        //
        // Wait for send completion.  Note that it is possible that the send
        // and receive completions get reversed here.  However, once in this loop,
        // we will always pick up two completions.
        //
        if( m_IoHelper.GetResults( &context, &cbXfer, __LINE__ ) == 0 )
        {
            if( bUseEvents )
            {
                HRESULT hr = m_pCq->Notify( ND_CQ_NOTIFY_ANY, &m_Ov );
                if( hr == ND_PENDING )
                {
                    hr = m_pCq->GetOverlappedResult( &m_Ov, TRUE );
                }
                if( FAILED( hr ) )
                {
                    printf( "IND2CompletionQueue::Notify failed with %08x\n", hr );
                    exit( __LINE__ );
                }
            }
            while( m_IoHelper.GetResults( &context, &cbXfer, __LINE__ ) == 0 )
            {
                NULL;
            }
        }

        if( context == &m_Results[RECV] && cbXfer != cbExpected )
        {
            printf( "Received %u bytes, expected %u\n",
                cbXfer, cbExpected );
            exit( __LINE__ );
        }

    } while( --iterations != 0 );
}


class NdPingPongClientBase : public NdPingPongBase
{
protected:
    Timer m_Timer;
    CpuMonitor m_Cpu;

public:
    NdPingPongClientBase()
    {
        printf( "Using %u processors. Sender Frequency is %I64d\n\n"
            " %9s %9s %9s %7s %11s\n",
            CpuMonitor::CpuCount(),
            Timer::Frequency(),
            "Size", "Iter", "Latency", "CPU", "Bytes/Sec"
            );
    }
    ~NdPingPongClientBase(){};

    void Connect( _In_ const sockaddr_in& v4Src, _In_ const sockaddr_in& v4Dst );
    void CompleteConnect();
};

void NdPingPongClientBase::Connect( _In_ const sockaddr_in& v4Src, _In_ const sockaddr_in& v4Dst )
{
    //
    // Connect to the server.
    //
    HRESULT hr = m_pConnector->Bind(
        reinterpret_cast<const sockaddr*>( &v4Src ),
        sizeof(v4Src)
        );
    if( FAILED( hr ) )
    {
        printf( "IND2Connector::Bind failed with %08x\n", hr );
        exit( __LINE__ );
    }

    hr = m_pConnector->Connect(
        m_pQp,
        reinterpret_cast<const sockaddr*>( &v4Dst ),
        sizeof(v4Dst),
        0,
        0,
        NULL,
        0,
        &m_Ov
        );
    if( hr == ND_PENDING )
    {
        hr = m_pConnector->GetOverlappedResult( &m_Ov, TRUE );
    }
    if( FAILED( hr ) || hr == ND_TIMEOUT )
    {
        printf( "IND2QueuePair::Connect failed with %08x\n", hr );
        exit( __LINE__ );
    }
}

void NdPingPongClientBase::CompleteConnect()
{
    //
    // Complete the connection - this transitions the endpoint so it can send.
    //
    HRESULT hr = m_pConnector->CompleteConnect( &m_Ov );
    if( hr == ND_PENDING )
    {
        hr = m_pConnector->GetOverlappedResult( &m_Ov, TRUE );
    }
    if( FAILED( hr ) || hr == ND_TIMEOUT )
    {
        printf( "IND2QueuePair::CompleteConnect failed with %08x\n", hr );
        exit( __LINE__ );
    }
}


template<class T>
class NdPingPongClient : public NdPingPongClientBase
{
    T m_IoHelper;

public:
    void RunTest(
        _In_ const struct sockaddr_in& v4Src,
        _In_ const struct sockaddr_in& v4Dst,
        _In_ DWORD queueDepth,
        _In_ DWORD nSge,
        _In_ bool bUseEvents
        );

    void Ping(
        _In_ DWORD nSge,
        _In_ DWORD iterations,
        _In_ bool bUseEvents
        );
};

template<class T>
void NdPingPongClient<T>::RunTest(
    _In_ const struct sockaddr_in& v4Src,
    _In_ const struct sockaddr_in& v4Dst,
    _In_ DWORD queueDepth,
    _In_ DWORD nSge,
    _In_ bool bUseEvents
    )
{
    IND2Adapter* pAdapter;
    T::OpenAdapter( v4Src, &pAdapter, __LINE__ );

    NdPingPongBase::Init( pAdapter, queueDepth, nSge );
    pAdapter->Release();

    m_IoHelper.Init( m_pQp, m_pCq, m_pMr, m_Buf, x_HdrLen + x_MaxXfer, nSge, __LINE__ );

    NdPingPongClientBase::Connect( v4Src, v4Dst );

    //
    // When using the SHIM, the INDEndpoint object in our IND2QueuePair object
    // changes in the Connect call.  Refresh our reference too.
    //
    m_IoHelper.RefreshQp( m_pQp, __LINE__ );

    //
    // Pre-post receive request - we keep one to prepost in Ping
    // so that we can start the loop with preposting before sending.
    //
    for( SIZE_T i = 0; i < queueDepth - 1; i++ )
    {
        m_IoHelper.Receive( &m_Results[RECV], __LINE__ );
    }

    NdPingPongClientBase::CompleteConnect();

    //
    // Warm up run, transfer only header.
    //
    m_IoHelper.SetSendSge( 0, m_Buf, x_HdrLen );

    Ping(
        nSge,
        1000,
        bUseEvents
        );

    for( DWORD szXfer = 1; szXfer <= x_MaxXfer; szXfer <<= 1 )
    {
        DWORD i;
        for( i = 0; i < nSge - 1; i++ )
        {
            m_IoHelper.SetSendSge( i, m_Buf + (x_HdrLen * i), x_HdrLen );
        }

        // Last SGE has the remainder of the data.
        m_IoHelper.SetSendSge( i, m_Buf + (x_HdrLen * i), x_HdrLen + szXfer - (x_HdrLen * i) );

        DWORD iterations = x_MaxIterations;
        if( iterations > (x_MaxVolume / szXfer) )
        {
            iterations = x_MaxVolume / szXfer;
        }

        m_Cpu.Start();
        m_Timer.Start();

        Ping(
            nSge,
            iterations,
            bUseEvents
            );

        m_Timer.End();
        m_Cpu.End();

        // Factor of 2 to account for ping *and* pong.
        double bytesSec = 2.0 * szXfer * iterations / (m_Timer.Report() / 1000000.0);
        // Factor of 2 to account for half-round trip latency.
        double latency = (m_Timer.Report() / iterations) / 2.0;
        printf(
            " %9u %9u %9.2f %7.2f %11.0f\n",
            szXfer,
            iterations,
            latency,
            m_Cpu.Report(),
            bytesSec
            );
    }

    NdPingPongBase::Shutdown();
}

template<class T>
void NdPingPongClient<T>::Ping(
    _In_ DWORD nSge,
    _In_ DWORD iterations,
    _In_ bool bUseEvents
    )
{
    DWORD cbExpected = 0;
    for( DWORD i = 0; i < nSge; i++ )
    {
        cbExpected += m_IoHelper.GetSendSgeLength(i);
    }
    ULONG flags = 0;
    if( cbExpected < m_IoHelper.MaxInline() )
    {
        flags |= ND_OP_FLAG_INLINE;
    }

    void* context;
    DWORD cbXfer;
    do
    {
        m_IoHelper.Receive( &m_Results[RECV], __LINE__ );
        m_IoHelper.Send( &m_Results[SEND], nSge, __LINE__, flags );

        //
        // Wait for send & receive completion.
        //
        for( int i = 0; i < 2; NULL )
        {
            if( m_IoHelper.GetResults( &context, &cbXfer, __LINE__ ) == 0 )
            {
                if( bUseEvents )
                {
                    HRESULT hr = m_pCq->Notify( ND_CQ_NOTIFY_ANY, &m_Ov );
                    if( hr == ND_PENDING )
                    {
                        hr = m_pCq->GetOverlappedResult( &m_Ov, TRUE );
                    }
                    if( FAILED( hr ) )
                    {
                        printf( "IND2CompletionQueue::Notify failed with %08x\n", hr );
                        exit( __LINE__ );
                    }
                }
                while( m_IoHelper.GetResults( &context, &cbXfer, __LINE__ ) == 0 )
                {
                    NULL;
                }
            }

            if( context == &m_Results[RECV] && cbXfer != cbExpected )
            {
                printf( "Received %u bytes, expected %u\n",
                    cbXfer, cbExpected );
                exit( __LINE__ );
            }
            i++;
        }

    } while( --iterations != 0 );
}


int __cdecl _tmain(int argc, TCHAR* argv[])
{
    bool bServer = false;
    bool bClient = false;
    DWORD nSge = 1;
    LONG queueDepth = 64;
    bool bPolling = false;
    bool bBlocking = false;
    int version = 1;
    struct sockaddr_in v4Server = {0};

    WSADATA wsaData;
    int ret = ::WSAStartup( MAKEWORD( 2, 2 ), &wsaData );
    if( ret != 0 )
    {
        printf( "Failed to initialize Windows Sockets: %d\n", ret );
        exit( __LINE__ );
    }

    for( int i = 1; i < argc; i++ )
    {
        const TCHAR* pArg;
        INT len;

        pArg = argv[i];

        // Skip leading dashes/slashes
        while( *pArg == '-' || *pArg == '/' )
        {
            pArg++;
        }

        switch( *pArg )
        {
        case _T('s'):
        case _T('S'):
            bServer = true;
            if( ++i == argc )
            {
                break;
            }

            len = sizeof( v4Server );
            WSAStringToAddress(
                argv[i],
                AF_INET,
                NULL,
                reinterpret_cast<struct sockaddr*>( &v4Server ),
                &len
                );
            break;
        case _T('c'):
        case _T('C'):
            bClient = true;
            if( ++i == argc )
            {
                break;
            }

            len = sizeof( v4Server );
            WSAStringToAddress(
                argv[i],
                AF_INET,
                NULL,
                reinterpret_cast<struct sockaddr*>( &v4Server ),
                &len
                );
            break;
        case _T('p'):
        case _T('P'):
            bPolling = true;
            nSge = _ttol( ++pArg );
            break;
        case _T('b'):
        case _T('B'):
            bBlocking = true;
            nSge = _ttol( ++pArg );
            break;
        case _T('l'):
        case _T('L'):
//
// Can't use freopen_s because it doesn't allow sharing.
// So supress the deprecated warning, because for our use
// it isn't deprecated.
//
#pragma warning( disable : 4996 )
            if( _tfreopen( ++pArg, _T("w"), stdout ) == NULL ||
                _tfreopen( pArg, _T("a+"), stderr ) == NULL )
#pragma warning( default : 4996 )
            {
                printf( "Could not open log file.\n" );
                exit( __LINE__ );
            }
            break;
        case _T('v'):
        case _T('V'):
            version = _ttol( ++pArg );
            break;
        case _T('q'):
        case _T('Q'):
            queueDepth = _ttol( ++pArg );
            break;

        default:
            _tprintf( _T("Unknown parameter %s\n"), pArg );
            ShowUsage();
            exit( __LINE__ );
        }
    }

    if( version != 1 && version != 2 )
    {
        printf( "Invalid version %d specified.\n\n", version );
        ShowUsage();
        exit( __LINE__ );
    }

    if( (bClient && bServer) || (!bClient && !bServer) )
    {
        printf( "Exactly one of client (c or "
            "server (s) must be specified.\n\n" );
        ShowUsage();
        exit( __LINE__ );
    }

    if( v4Server.sin_addr.s_addr == 0 )
    {
        printf( "Bad address.\n\n");
        ShowUsage();
        exit( __LINE__ );
    }

    if( v4Server.sin_port == 0 )
    {
        v4Server.sin_port = htons( x_DefaultPort );
    }

    if( (bPolling && bBlocking) )
    {
        printf( "Exactly one of blocking (b or "
            "polling (p) must be specified.\n\n" );
        ShowUsage();
        exit( __LINE__ );
    }

    if( nSge == 0 )
    {
        printf( "Invalid or missing SGE, expected positive integer after p or b.\n\n" );
        ShowUsage();
        exit( __LINE__ );
    }

    OVERLAPPED Ov;
    Ov.hEvent = CreateEvent( NULL, FALSE, FALSE, NULL );
    if( Ov.hEvent == NULL )
    {
        printf( "Failed to allocate event for overlapped operations.\n" );
        exit( __LINE__ );
    }

    HRESULT hr = NdStartup();
    if( FAILED(hr) )
    {
        printf( "NdStartup failed with %08x\n", hr );
        exit( __LINE__ );
    }

    if( bServer )
    {
        if( version == 1 )
        {
            NdPingPongServer<NDIoHelperV1> server;
            server.RunTest( v4Server, queueDepth, nSge, bBlocking );
        }
        else
        {
            NdPingPongServer<NDIoHelperV2> server;
            server.RunTest( v4Server, queueDepth, nSge, bBlocking );
        }
    }
    else
    {
        struct sockaddr_in v4Src;

        SIZE_T len = sizeof(v4Src);
        hr = NdResolveAddress(
            (const struct sockaddr*)&v4Server,
            sizeof(v4Server),
            (struct sockaddr*)&v4Src,
            &len );
        if( FAILED( hr ) )
        {
            printf( "NdResolveAddress failed with %08x\n", hr );
            exit( __LINE__ );
        }

        if( version == 1 )
        {
            NdPingPongClient<NDIoHelperV1> client;
            client.RunTest( v4Src, v4Server, queueDepth, nSge, bBlocking );
        }
        else
        {
            NdPingPongClient<NDIoHelperV2> client;
            client.RunTest( v4Src, v4Server, queueDepth, nSge, bBlocking );
        }
    }

    hr = NdCleanup();
    if( FAILED(hr) )
    {
        printf( "NdCleanup failed with %08x\n", hr );
        exit( __LINE__ );
    }

    CloseHandle( Ov.hEvent );

    _fcloseall();
    WSACleanup();

    WSACleanup();
    return 0;
}
